/*
############################################################################
# This file contains the word functions for the pfth Forth interpreter.
#
# Copyright (c) 2012 Dave Hein
# MIT Licensed
############################################################################
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include "pfth.h"
#include "words.h"

char pad[256];
char *printptr = pad + 255;

int basefunc(void *parm)
{
    push((int)&base);
    return errorcode;
}

int FAST fetchfunc(void *parm)
{
    if (stkidx < 1)
      return RETURN_ERROR;
    stack[stkidx-1] = *(int *)stack[stkidx-1];
    return errorcode;
}

int FAST storefunc(void *parm)
{
    int *ptr;
    if (stkidx < 2)
      return RETURN_ERROR;
    ptr = (int *)stack[--stkidx];
    *ptr = stack[--stkidx];
    return errorcode;
}

int FAST cfetchfunc(void *parm)
{
    if (stkidx < 1)
      return RETURN_ERROR;
    stack[stkidx-1] = *(unsigned char *)stack[stkidx-1];
    return errorcode;
}

int FAST cstorefunc(void *parm)
{
    unsigned char *ptr;
    if (stkidx < 2)
      return RETURN_ERROR;
    ptr = (unsigned char *)stack[--stkidx];
    *ptr = stack[--stkidx];
    return errorcode;
}

int FAST plusfunc(void *parm)
{
    if (stkidx < 2) return RETURN_ERROR;
    stkidx--;
    stack[stkidx-1] = stack[stkidx-1] + stack[stkidx];
    return errorcode;
}

int FAST minusfunc(void *parm)
{
    int value = pop();
    push(pop() - value);
    return errorcode;
}

int multfunc(void *parm)
{
    push(pop() * pop());
    return errorcode;
}

int dividefunc(void *parm)
{
    int value = pop();
    push(pop() / value);
    return errorcode;
}

int FAST ExecuteBody(int *body)
{
    dictentryT *entry;
    int retval = errorcode;
    int evalsave = evaluate;
    evaluate = 0;
    retstk[retidx++] = (int)execaddr;
    execaddr = body;
    while ((entry = (dictentryT *)(*execaddr)))
    {
#if DEBUG
        printf("execfunc ");
        Type(entry->name, entry->len);
        printf("\n");
#endif
        if ((retval = EXECUTE(entry)) <= RETURN_EXIT)
            break;
        execaddr++;
    }
    execaddr = (int *)retstk[--retidx];
    evaluate = evalsave;
    return retval;
}

int FAST execfunc(void *parm)
{
    dictentryT *entry = (dictentryT *)parm;
    return ExecuteBody((int *)ALIGN(entry->name + entry->len));
}

int addrfunc(void *parm)
{
    dictentryT *entry = (dictentryT *)parm;
    push((int)ALIGN(entry->name + entry->len));
    if (entry->does) ExecuteBody(entry->does);
    return errorcode;
}

int createfunc(void *parm)
{
    int len;
    char *wordptr = Word(&len);
    Create(wordptr, len, addrfunc);
    return errorcode;
}

int refillfunc(void *parm)
{
    push(Refill());
    return errorcode;
}

void MakeEntry(char *name, wordfuncP wordfunc, int flags)
{
    Create(name, strlen(name), wordfunc);
    lastword->flags = flags;
}

int colonfunc(void *parm)
{
    int len;
    char *wordptr = Word(&len);
    Create(wordptr, len, execfunc);
    compileflag = -1;
    return errorcode;
}

int semicolonfunc(void *parm)
{
    compileflag = 0;
    here = ALIGN(here);
    *((int *)here) = 0;
    here += 4;
    return errorcode;
}

int includefunc(void *parm)
{
    int len;
    char fname[100];
    char *wordptr = Word(&len);

    memcpy(fname, wordptr, len);
    fname[len] = 0;

    Include(fname);
    return errorcode;
}

int evaluatefunc(void *parm)
{
    int len = pop();
    char *ptr = (char *)pop();

    instack[inidx++] = (char *)inputfile;
    instack[inidx++] = inputptr;
    instack[inidx++] = (char *)inputidx;
    instack[inidx++] = (char *)inputlen;

    inputfile = 0;
    inputptr = ptr;
    inputlen = len;
    inputidx = 0;

    //Evaluate(ptr, len);
    Evaluate();

    inputlen  = (int)instack[--inidx];
    inputidx  = (int)instack[--inidx];
    inputptr  = instack[--inidx];
    inputfile = (FILE *)instack[--inidx];

    return errorcode;
}

int FAST toretfunc(void *parm)
{
    pushret(pop());
    return errorcode;
}

int FAST fromretfunc(void *parm)
{
    push(popret());
    return errorcode;
}

int FAST dropfunc(void *parm)
{
    pop();
    return errorcode;
}

int FAST pickfunc(void *parm)
{
    int index;
    if (stkidx < 2) return RETURN_ERROR;
    index = stack[stkidx-1];
    if (stkidx < index+2) return RETURN_ERROR;
    stack[stkidx-1] = stack[stkidx-index-2];
    return errorcode;
}

int FAST rollfunc(void *parm)
{
    int index, start, end;
    if (stkidx < 2) return RETURN_ERROR;
    index = stack[stkidx-1];
    if (stkidx < index+2) return RETURN_ERROR;
    stack[stkidx-1] = stack[stkidx-index-2];
    start = stkidx - index - 2;
    end = stkidx - 2;
    for (index = start; index <= end; index++)
        stack[index] = stack[index+1];
    stkidx--;
    return errorcode;
}

int equalfunc(void *parm)
{
    int value2 = pop();
    int value1 = pop();
    push(-(value1 == value2));
    return errorcode;
}

int FAST lessfunc(void *parm)
{
    int value1, value2;
    if (stkidx < 2) return RETURN_ERROR;
    value2 = stack[--stkidx];
    value1 = stack[stkidx-1];
    stack[stkidx-1] = -(value1 < value2);
    return errorcode;
}

int greaterfunc(void *parm)
{
    int value2 = pop();
    int value1 = pop();
    push(-(value1 > value2));
    return errorcode;
}

int FAST andfunc(void *parm)
{
    if (stkidx < 2) return RETURN_ERROR;
    stkidx--;
    stack[stkidx-1] = stack[stkidx-1] & stack[stkidx];
    return errorcode;
}

int orfunc(void *parm)
{
    push(pop() | pop());
    return errorcode;
}

int xorfunc(void *parm)
{
    push(pop() ^ pop());
    return errorcode;
}

int lshiftfunc(void *parm)
{
    unsigned int shift = pop();
    unsigned int value = pop();
    push(value << shift);
    return errorcode;
}

int rshiftfunc(void *parm)
{
    unsigned int shift = pop();
    unsigned int value = pop();
    push(value >> shift);
    return errorcode;
}

int depthfunc(void *parm)
{
    push(stkidx);
    return errorcode;
}

int emitfunc(void *parm)
{
    putchar(pop());
    fflush(stdout);
    return errorcode;
}

#if 0
int leavefunc(void *parm)
{
    if (compileflag && evaluate)
    {
        int index;
        *(int *)here = (int)parm;
        here += 4;
        for (index = retidx++; index > 0; index--)
        {
            if (((int)retstk[index-1]) & 3)
            {
                retstk[index] = (int)(here + 3);
                break;
            }
            retstk[index] = retstk[index-1];
        }
        here += 4;
    }
    else
    {
        popret();
        popret();
        execaddr = (int *)execaddr[1];
    }
    return errorcode;
}

int FAST loopfunc(void *parm)
{
    if (compileflag && evaluate)
    {
        int flag;
        int *ptr;
        *(int *)here = (int)parm;
        here += 4;
        while (1)
        {
            flag = (int)retstk[--retidx];
            ptr = (int *)(flag & ~3);
            flag &= 3;
            if (flag > 1) *ptr = (int)here;
            if (flag != 3) break;
        }
        *(int *)here = (int)ptr;
        here += 4;
    }
    else
    {
        int index, limit;
        
        if (retidx < 2)
            return RETURN_ERROR;

        index = retstk[retidx-1];
        limit = retstk[retidx-2];
        
        if (++index < limit)
        {
            execaddr = (int *)execaddr[1];
            retstk[retidx-1] = index;
        }
        else
        {
            execaddr++;
            retidx -= 2; 
        }       
    }
    return errorcode;
}

int FAST plusloopfunc(void *parm)
{
    if (compileflag && evaluate)
    {
        int flag;
        int *ptr;
        *(int *)here = (int)parm;
        here += 4;
        while (1)
        {
            flag = (int)retstk[--retidx];
            ptr = (int *)(flag & ~3);
            flag &= 3;
            if (flag > 1) *ptr = (int)here;
            if (flag != 3) break;
        }
        *(int *)here = (int)ptr;
        here += 4;
    }
    else
    {
        int index = popret();
        int limit = popret();

        index += pop();

        if (index < limit)
        {
            execaddr = (int *)execaddr[1];
            pushret(limit);
            pushret(index);
        }
        else
            execaddr++;
    }
    return errorcode;
}
#endif

int FAST _litfunc(void *parm)
{
    execaddr++;
    stack[stkidx++] = *execaddr;
    return errorcode;
}

int byefunc(void *parm)
{
    return RETURN_BYE;
}

int unusedfunc(void *parm)
{
    push((int)(((char *)&codespace[CODE_SIZE]) - here));
    return errorcode;
}

int tibfunc(void *parm)
{
    push((int)inputptr);
    return errorcode;
}

int lastfunc(void *parm)
{
    push((int)&lastword);
    return errorcode;
}

int dpfunc(void *parm)
{
    push((int)&here);
    return errorcode;
}

int numbertibfunc(void *parm)
{
    push((int)&inputlen);
    return errorcode;
}

int statefunc(void *parm)
{
    push((int)&compileflag);
    return errorcode;
}

int abortfunc(void *parm)
{
    stkidx = 0;
    retidx = 0;
    return RETURN_ERROR;
}

int quitfunc(void *parm)
{
    retidx = 0;
    return RETURN_QUIT;
}

int exitfunc(void *parm)
{
    return RETURN_EXIT;
}

int infunc(void *parm)
{
    push((int)&inputidx);
    return errorcode;
}

int wordfunc(void *parm)
{
    int len;
    int delimiter = pop();
    char *wordptr = WordDelimiter(&len, delimiter);
    pad[0] = len;
    memcpy(pad+1, wordptr, len);
    push((int)pad);
    return errorcode;
}

int executefunc(void *parm)
{
    int retval;
    dictentryT *entry = (dictentryT *)pop();
    int evalsave = evaluate;
    evaluate = 0;
    if (errorcode == RETURN_ERROR) return errorcode;
    retval = EXECUTE(entry);
    evaluate = evalsave;
    return retval;
}

int acceptfunc(void *parm)
{
    int num = pop();
    char *ptr = (char *)pop();
    if (errorcode == RETURN_ERROR) return errorcode;
    if (fgets(ptr, num, stdin) > 0)
    {
        RemoveCRLF(ptr);
        push(strlen(ptr));
    }
    else
        push(0);
    return errorcode;
}

int keyfunc(void *parm)
{
    int value;
    while ((value = getch()) == -1);
    push(value);
    return errorcode;
}

int findfunc(void *parm)
{
    char *ptr = (char *)pop();
    int num = *ptr;
    dictentryT *entry = FindWord(ptr+1, num);
    if (!entry)
    {
        push((int)ptr);
        push(0);
        return errorcode;
    }
    push((int)entry);
    if (entry->flags & FLAG_IMMEDIATE)
        push(1);
    else
        push(-1);
    return errorcode;
}

int digit2ascii(int digit)
{
    if (digit < 10)
        digit += '0';
    else
        digit += 'A' - 10;
    return digit;
}

int ConvertUnsignedNumber(unsigned int value)
{
    printptr = pad + 255;
    *printptr-- = digit2ascii(value % base);
    while ((value /= base))
        *printptr-- = digit2ascii(value % base);
    return (int)(pad + 255 - printptr);
}

int ConvertSignedNumber(int value)
{
    if (value >= 0)
        return ConvertUnsignedNumber((unsigned int)value);
    ConvertUnsignedNumber((unsigned int)(-value));
    *printptr-- = '-';
    return (int)(pad + 255 - printptr);
}

void PrintBody(void *parm)
{
    dictentryT *entry;

    while (*(int *)parm)
    {
        entry = (dictentryT *)(*(int *)parm);
        Type(entry->name, entry->len);
        printf(" ");
        parm += 4;
        if (entry->flags & FLAG_LITERAL)
        {
            int len;
            int value = *(int *)parm;
            if (entry->flags & FLAG_JUMP) value -= (int)parm;
            len = ConvertSignedNumber(value);
            Type(printptr+1, len);
            printf(" ");
            parm += 4;
        }
        else if (entry->flags & FLAG_STRING)
        {
            int num = *(char *)parm++;
            Type(parm, num);
            printf("\" ");
            parm += num;
            parm = ALIGN(parm);
        }
    }
}

int seefunc(void *parm)
{
    int len;
    char *wordptr = Word(&len);
    dictentryT *wordentry = FindWord(wordptr, len);

    if (!wordentry)
    {
        Type(wordptr, len);
        printf(" ?\n");
        return RETURN_ERROR;
    }

    if (wordentry->func == execfunc)
    {
        printf(": ");
        Type(wordentry->name, wordentry->len);
        printf(" ");
        parm = ALIGN(wordentry->name + wordentry->len);
        PrintBody(parm);
        printf(";");
        if (wordentry->flags & FLAG_IMMEDIATE)
            printf(" immediate");
    }
    else if (wordentry->func == addrfunc)
    {
        int len;
        int *ptr = (int *)ALIGN(wordentry->name + wordentry->len);
        len = ConvertSignedNumber(ptr[0]);
        printf("create ");
        Type(wordentry->name, wordentry->len);
        printf(" ");
        Type(printptr+1, len);
        printf(" ");
        len = ConvertSignedNumber(ptr[1]);
        Type(printptr+1, len);
        printf(" ");
        if (wordentry->does)
        {
            printf("does> ");
            PrintBody(wordentry->does);
        }
    }
    else
    {
        printf("kernel ");
        Type(wordentry->name, wordentry->len);
    }
    printf("\n");

    return errorcode;
}

int FAST modfunc(void *parm)
{
    int value1, value2;
    if (stkidx < 2) return RETURN_ERROR;
    value2 = stack[--stkidx];
    value1 = stack[stkidx-1];
    stack[stkidx-1] = value1 % value2;
    return errorcode;
}

void InitializeBasicWords()
{
    MakeEntry(":", colonfunc, 0);
    MakeEntry(";", semicolonfunc, FLAG_IMMEDIATE);
    MakeEntry("create", createfunc, 0);
    MakeEntry("refill", refillfunc, 0);

    MakeEntry("_lit", _litfunc, FLAG_LITERAL);
    _litword = lastword;
    MakeEntry("@", fetchfunc, 0);
    MakeEntry("!", storefunc, 0);
    storeword = lastword;
    MakeEntry("c!", cstorefunc, 0);
    MakeEntry("c@", cfetchfunc, 0);

    MakeEntry("+", plusfunc, 0);
    MakeEntry("-", minusfunc, 0);
    MakeEntry("*", multfunc, 0);
    MakeEntry("/", dividefunc, 0);
    MakeEntry("mod", modfunc, 0);

    MakeEntry("=", equalfunc, 0);
    MakeEntry("<", lessfunc, 0);
    MakeEntry(">", greaterfunc, 0);

    MakeEntry("and", andfunc, 0);
    MakeEntry("or", orfunc, 0);
    MakeEntry("xor", xorfunc, 0);

    MakeEntry("rshift", rshiftfunc, 0);
    MakeEntry("lshift", lshiftfunc, 0);

    MakeEntry("depth", depthfunc, 0);
    MakeEntry("pick", pickfunc, 0);
    MakeEntry("roll", rollfunc, 0);
    MakeEntry("drop", dropfunc, 0);
    MakeEntry(">r", toretfunc, 0);
    MakeEntry("r>", fromretfunc, 0);

    MakeEntry("evaluate", evaluatefunc, 0);
    MakeEntry("word", wordfunc, 0);
    MakeEntry("find", findfunc, 0);
    MakeEntry("execute", executefunc, 0);

#if 0
    MakeEntry("leave", leavefunc, FLAG_IMMEDIATE | FLAG_LITERAL | FLAG_JUMP);
    MakeEntry("loop", loopfunc, FLAG_IMMEDIATE | FLAG_LITERAL | FLAG_JUMP);
    MakeEntry("+loop", plusloopfunc, FLAG_IMMEDIATE | FLAG_LITERAL | FLAG_JUMP);
#endif

    MakeEntry("abort", abortfunc, 0);
    MakeEntry("quit", quitfunc, 0);
    MakeEntry("exit", exitfunc, 0);

    MakeEntry("base", basefunc, 0);
    MakeEntry("state", statefunc, 0);
    MakeEntry("tib", tibfunc, 0);
    MakeEntry("#tib", numbertibfunc, 0);
    MakeEntry(">in", infunc, 0);
    MakeEntry("last", lastfunc, 0);
    MakeEntry("dp", dpfunc, 0);

    MakeEntry("accept", acceptfunc, 0);
    MakeEntry("key", keyfunc, 0);
    MakeEntry("emit", emitfunc, 0);

    MakeEntry("see", seefunc, 0);
    MakeEntry("include", includefunc, 0);
    MakeEntry("bye", byefunc, 0);
    MakeEntry("unused", unusedfunc, 0);
}
/*
+--------------------------------------------------------------------
|  TERMS OF USE: MIT License
+--------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files
(the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
+------------------------------------------------------------------
*/
